/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "xunfeivisionengine_p.h"
#include <unistd.h>
#include <map>
#include "imageloader.h"
#include "servererror.h"
#include "token.h"
#include "util.h"

namespace {
std::map<int, std::string> imageStyleMap = {
    {0, "探索无限"},    {1, "古风"},     {2, "二次元"},      {3, "写实风格"},
    {4, "浮世绘"},      {5, "low poly"}, {6, "未来主义"},    {7, "像素风格"},
    {8, "概念艺术"},    {9, "赛博朋克"}, {10, "洛丽塔风格"}, {11, "巴洛克风格"},
    {12, "超现实主义"}, {13, "水彩画"},  {14, "蒸汽波艺术"}, {15, "油画"},
    {16, "卡通画"},
};
}

XunfeiVisionEnginePrivate::XunfeiVisionEnginePrivate() {}

XunfeiVisionEnginePrivate::~XunfeiVisionEnginePrivate() {}

std::string XunfeiVisionEnginePrivate::engineName() {
    return std::string("xunfei");
}

void XunfeiVisionEnginePrivate::setAppId(const std::string &appId) {
    appId_ = appId;
}

void XunfeiVisionEnginePrivate::setApiKey(const std::string &apiKey) {
    apiKey_ = apiKey;
}

void XunfeiVisionEnginePrivate::setSecretKey(const std::string &secretKey) {
    secretKey_ = secretKey;
}

void XunfeiVisionEnginePrivate::setPrompt2ImageCallback(
    ai_engine::lm::vision::ImageResultCallback callback) {
    imageResultCallback_ = callback;
}

void XunfeiVisionEnginePrivate::setPrompt2ImageSize(int width, int height) {
    imageWidth_ = width;
    imageHeight_ = height;
}

void XunfeiVisionEnginePrivate::setPrompt2ImageNumber(int number) {
    imageNumber_ = number;
}

void XunfeiVisionEnginePrivate::setPrompt2ImageStyle(int style) {
    imageStyle_ = imageStyleMap[style];
}

bool XunfeiVisionEnginePrivate::prompt2Image(
    const std::string &prompt, ai_engine::lm::EngineError &error) {
    isStopped_ = false;

    if (prompt.empty()) {
        error = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Vision,
            ai_engine::lm::EngineErrorCategory::Initialization,
            (int)ai_engine::lm::VisionEngineErrorCode::TooLittleData,
            "文生图提示词为空");
        return false;
    }
    if (!imageResultCallback_) {
        error = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Vision,
            ai_engine::lm::EngineErrorCategory::Initialization,
            (int)ai_engine::lm::VisionEngineErrorCode::TooLittleParams,
            "文生图未设置回调函数");
        return false;
    }

    Json::Value bodyJson;
    bodyJson["header"]["app_id"] = appId_;
    bodyJson["parameter"]["chat"]["domain"] = "general";
    bodyJson["parameter"]["chat"]["width"] = imageWidth_;
    bodyJson["parameter"]["chat"]["height"] = imageHeight_;
    bodyJson["payload"]["message"]["text"][0]["role"] = "user";
    if (imageStyle_.empty()) {
        bodyJson["payload"]["message"]["text"][0]["content"] =
            std::string(prompt);
    } else if (imageStyle_ == "low poly") {
        bodyJson["payload"]["message"]["text"][0]["content"] =
            std::string(prompt) + "，" + "低多边形";
    } else {
        bodyJson["payload"]["message"]["text"][0]["content"] =
            std::string(prompt) + "，" + imageStyle_;
    }
    std::string bodyString = bodyJson.toStyledString();

    cpr::Url url{token::getAuthenticationUrl(apiKey_, secretKey_, hostUrl_)};
    cpr::Header header{cpr::Header{{"Content-Type", "application/json"},
                                   {"charset", "UTF-8"}}};
    cpr::Body body{cpr::Body{{bodyString}}};
    cpr::Response response = cpr::Post(url, header, body);

    return processPrompt2ImageResult(response, error);
}

bool XunfeiVisionEnginePrivate::processPrompt2ImageResult(
    cpr::Response &response, ai_engine::lm::EngineError &error) {
    if (isStopped_) {
        return true;
    }

    if (response.error) {
        error = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Vision,
            ai_engine::lm::EngineErrorCategory::OutputError,
            (int)ai_engine::lm::VisionEngineErrorCode::NetworkDisconnected,
            response.error.message);
        fprintf(stderr, "net error: %s\n", response.error.message.c_str());
        return false;
    }

    if (int errorCode = server_error::parseErrorCode(response.text)) {
        auto erroraTuple = server_error::errorCode2visionResult(errorCode);
        error = ai_engine::lm::EngineError(
            ai_engine::lm::AiCapability::Vision, std::get<0>(erroraTuple),
            (int)std::get<1>(erroraTuple), response.text);
        fprintf(stderr, "xunfei text2image failed: %s\n",
                response.text.c_str());
        return false;
    }

    Json::Value resultJson = util::formatJsonFromString(response.text);
    ai_engine::lm::vision::ImageData imageData;
    for (int i = 0; i < resultJson["payload"]["choices"]["text"].size(); i++) {
        auto imageDataString = util::base64Decode(
            resultJson["payload"]["choices"]["text"][i]["content"].asString());
        auto imageInfo = vision::image_loader::getImageInfoFromData(
            (unsigned char *)imageDataString.data(), imageDataString.size());
        ai_engine::lm::EngineError error;
        imageData.error = error;
        imageData.format = imageInfo.format;
        imageData.width = imageInfo.width;
        imageData.height = imageInfo.height;
        imageData.imageData =
            std::vector<char>(imageDataString.begin(), imageDataString.end());
        imageData.index = i;
        imageData.total = resultJson["payload"]["choices"]["text"].size();
    }
    imageResultCallback_(imageData);
    return true;
}